#!/usr/bin/env python

from argparse import ArgumentParser, SUPPRESS
from collections import OrderedDict
from shutil import rmtree
import os
import re
import sys

from ._version import __version__
from .exceptions import MissingGenomeConfigError, MissingFolderError
from .asset_build_packages import *

import logmuse
import pypiper
import refgenconf
from refgenconf import RefGenConf, MissingAssetError, MissingGenomeError
from refgenconf.const import *
from ubiquerg import is_url, query_yes_no
import yacman

_LOGGER = None

BUILD_CMD = "build"
INIT_CMD = "init"
PULL_CMD = "pull"
LIST_LOCAL_CMD = "list"
LIST_REMOTE_CMD = "listr"
GET_ASSET_CMD = "seek"
INSERT_CMD = "add"
REMOVE_CMD = "remove"

BUILD_SPECIFIC_ARGS = ('fasta', 'gtf', 'gff', 'context', 'refgene')

# This establishes the API with the server
refgenie_server_api = {
    "list_available_genomes": "/genomes",
    'list_assets_by_genome': "/genome/{genome}",
    'download_asset': "/asset/{genome}/{asset}",
}


class _VersionInHelpParser(ArgumentParser):
    def format_help(self):
        """ Add version information to help text. """
        return "version: {}\n".format(__version__) + \
               super(_VersionInHelpParser, self).format_help()


def build_argparser():
    """
    Builds argument parser.

    :return argparse.ArgumentParser
    """

    banner = "%(prog)s - builds and manages reference genome assemblies"
    additional_description = "\nhttps://refgenie.databio.org"

    parser = _VersionInHelpParser(
        description=banner,
        epilog=additional_description)

    parser.add_argument(
        "-V", "--version",
        action="version",
        version="%(prog)s {v}".format(v=__version__))

    subparsers = parser.add_subparsers(dest="command")

    def add_subparser(cmd, description):
        return subparsers.add_parser(
            cmd, description=description, help=description)

    subparser_messages = {
        INIT_CMD: "Initialize a genome configuration.",
        LIST_LOCAL_CMD: "List available local assets.",
        LIST_REMOTE_CMD: "List available remote assets.",
        PULL_CMD: "Download assets.",
        BUILD_CMD: "Build genome assets.",
        GET_ASSET_CMD: "Get the path to a local asset.",
        INSERT_CMD: "Add local asset to the config file.",
        REMOVE_CMD: "Remove a local asset."
    }

    sps = {}
    for cmd, desc in subparser_messages.items():
        sps[cmd] = add_subparser(cmd, desc)
        # It's required for init
        sps[cmd].add_argument(
            '-c', '--genome-config', required=(cmd == INIT_CMD), dest="genome_config",
            help="Path to local genome configuration file.")

    sps[INIT_CMD].add_argument('-s', '--genome-server', default=DEFAULT_SERVER,
                               help="URL to use for the genome_server attribute in config file."
                                    " Defaults : {}".format(DEFAULT_SERVER))
    sps[BUILD_CMD] = pypiper.add_pypiper_args(
        sps[BUILD_CMD], groups=None, args=["recover", "config", "new-start"])

    # Add any arguments specific to subcommands.

    sps[BUILD_CMD].add_argument(
        "-d", "--docker", action="store_true",
        help="Run all commands in the refgenie docker container.")

    sps[BUILD_CMD].add_argument(
        '-v', '--volumes', nargs="+", required=False, default=None,
        help='If using docker, also mount these folders as volumes')

    sps[BUILD_CMD].add_argument(
        '-o', '--outfolder', dest='outfolder', required=False, default=None,
        help='Override the default path to genomes folder, which is the '
             'genome_folder attribute in the genome configuration file.')

    for cmd in [PULL_CMD, GET_ASSET_CMD, BUILD_CMD, INSERT_CMD, LIST_LOCAL_CMD, LIST_REMOTE_CMD, REMOVE_CMD]:
        # genome is not required for listing actions
        sps[cmd].add_argument(
            "-g", "--genome", required=cmd not in (LIST_REMOTE_CMD, LIST_LOCAL_CMD),
            help="Reference assembly ID, e.g. mm10")
    for cmd in [PULL_CMD, GET_ASSET_CMD, BUILD_CMD, INSERT_CMD, REMOVE_CMD]:
        sps[cmd].add_argument(
            "-a", "--asset", required=not cmd == REMOVE_CMD, nargs='+',
            help="Name of one or more assets (keys in genome config file)")

    sps[PULL_CMD].add_argument(
        "-u", "--no-untar", action="store_true",
        help="Do not extract tarballs.")

    sps[INSERT_CMD].add_argument(
        "-p", "--path", required=True,
        help="Relative path to asset")

    # Finally, arguments to the build command to give the files needed to do
    # the building. These should eventually move to a more flexible system that
    # doesn't require them to be hard-coded here in order to be recognized

    for arg in BUILD_SPECIFIC_ARGS:
        sps[BUILD_CMD].add_argument(
            "--{arg}".format(arg=arg), required=False, help=SUPPRESS)

    return parser


def copy_or_download_file(input_string, outfolder):
    """
    Given an input file, which can be a local file or a URL, and output folder,
    this downloads or copies the file into the output folder.

    :param str input_string: Can be either a URL or a path to a local file
    :param str outfolder: Where to store the result.
    :return str, str: output/result file and command
    """
    result_file = os.path.join(outfolder, os.path.basename(input_string))
    parts = ["wget -O", result_file, input_string] \
        if is_url(input_string) else ["cp", input_string, result_file]
    return result_file, " ".join(parts)


def convert_file(input_fasta, output_file, conversions):
    """
    Given an input file, output file, and a list of conversions, gives the appropriate output file.

    :param str output_file: Path to local output file you want to create
    :param dict conversions: A dictionary of shell commands to convert files of a given type.
    """
    form = {"INPUT": input_fasta, "OUTPUT": output_file}
    _, ext = os.path.splitext(input_fasta)
    if ext in conversions:
        return conversions[ext].format(**form)


def default_config_file():
    """
    Path to default compute environment settings file.

    :return str: Path to default compute settings file
    """
    return os.path.join(os.path.dirname(__file__), "refgenie.yaml")


def get_asset_vars(genome, asset_key, outfolder, specific_args=None):
    """
    Gives a dict with variables used to populate an asset path.
    """
    asset_outfolder = os.path.join(outfolder, asset_key)
    asset_vars = {"genome": genome,
                  "asset": asset_key,
                  "asset_outfolder": asset_outfolder}
    if specific_args:
        asset_vars.update(specific_args)
    return asset_vars


def refgenie_add(rgc, args):
    outfolder = os.path.abspath(os.path.join(rgc.genome_folder, args.genome))
    asset_vars = get_asset_vars(args.genome, args.asset, outfolder)
    rgc.update_assets(args.genome, args.asset, {CFG_ASSET_PATH_KEY: args.path.format(**asset_vars)})
    # Write the updated refgenie genome configuration
    rgc.write()


def refgenie_build(rgc, args):
    """
    Runs the refgenie build recipe.

    :param refgenconf.RefGenConf rgc: genome configuration instance
    :param argparse.Namespace args: parsed command-line options/arguments
    """

    # Build specific args
    specific_args = {k: getattr(args, k) for k in BUILD_SPECIFIC_ARGS}

    if args.genome:
        genome = args.genome
    else:
        # This can probably be eliminated now that with flexible building
        genome = os.path.basename(args.input)
        # eliminate extensions to get canonical genome name.
        for strike in [".fasta.gz$", ".fa.gz$", ".fasta$", ".fa$", ".gz$", ".2bit$"]:
            genome = re.sub(strike, "", genome)

    _LOGGER.info("Using genome name: {}".format(genome))

    if not hasattr(args, "outfolder") or not args.outfolder:
        # Default to genome_folder
        _LOGGER.debug("No outfolder provided, using genome config.")
        args.outfolder = rgc.genome_folder

    outfolder = os.path.abspath(os.path.join(args.outfolder, genome))
    if not _writeable(outfolder):
        _LOGGER.error("Insufficient permissions to write to output folder: {}".
                      format(outfolder))
        return

    _LOGGER.info("Output to: {} {} {}".format(genome, args.outfolder, outfolder))
    _LOGGER.debug("Default config file: {}".format(default_config_file()))

    if args.config_file and not os.path.isfile(args.config_file):
        _LOGGER.debug("Config file path isn't a file: {}".
                      format(args.config_file))
        args.config_file = default_config_file()

    def build_asset(genome, asset_key, asset_build_package, outfolder, specific_args):
        """
        Builds assets with pypiper and updates a genome config file.

        This function actually run the build commands in a given build package,
        and then update the refgenie config file.

        :param str genome: The assembly key; e.g. 'mm10'.
        :param str asset_key: The unique asset identifier; e.g. 'bowtie2_index'
        :param dict asset_build_package: A dict (see examples) specifying lists
            of required inputs, commands to run, and outputs to register as
            assets.
        """
        _LOGGER.debug("Asset build package: " + str(asset_build_package))
        asset_vars = get_asset_vars(genome, asset_key, outfolder, specific_args)
        asset_outfolder = os.path.join(outfolder, asset_key)

        _LOGGER.debug(str([x.format(**asset_vars) for x in asset_build_package[CMD_LST]]))

        tk.make_dir(asset_outfolder)
        target = os.path.join(asset_outfolder, "build_complete.flag")
        command_list_populated = [x.format(**asset_vars) for x in asset_build_package[CMD_LST]]

        touch_target = "touch {target}".format(target=target)
        command_list_populated.append(touch_target)

        _LOGGER.debug("Command list populated: " + str(command_list_populated))

        pm.run(command_list_populated, target, container=pm.container)

        for asset in asset_build_package[ASSETS].keys():
            rgc.update_assets(genome, asset, {
                CFG_ASSET_PATH_KEY: asset_build_package[ASSETS][asset][PTH].format(**asset_vars),
                CFG_ASSET_DESC_KEY: asset_build_package[ASSETS][asset][ASSET_DESC]
            })

        # Write the updated refgenie genome configuration
        rgc.write()

    pm = pypiper.PipelineManager(name="refgenie", outfolder=outfolder, args=args)
    tk = pypiper.NGSTk(pm=pm)

    if args.docker:
        # Set up some docker stuff
        if args.volumes:
            volumes = volumes.append(outfolder)
        else:
            volumes = outfolder

    for asset_key in args.asset:
        if asset_key in asset_build_packages.keys():
            asset_build_package = asset_build_packages[asset_key]
            _LOGGER.debug(specific_args)
            required_inputs = ", ".join(asset_build_package[REQ_IN])
            _LOGGER.info("Inputs required to build '{}': {}".format(asset_key, required_inputs))
            for required_input in asset_build_package[REQ_IN]:
                if not specific_args[required_input]:
                    raise ValueError(
                        "Argument '{}' is required to build asset '{}', but not provided".format(required_input,
                                                                                                 asset_key))

            for required_asset in asset_build_package[REQ_ASSETS]:
                try:
                    if not rgc.get_asset(args.genome, required_asset):
                        raise ValueError(
                            "Asset '{}' is required to build asset '{}', but not provided".format(required_asset,
                                                                                                  asset_key))
                except refgenconf.exceptions.MissingGenomeError:
                    raise ValueError(
                        "Asset '{}' is required to build asset '{}', but not provided".format(required_asset,
                                                                                              asset_key))
            if args.docker:
                pm.get_container(asset_build_package[CONT], volumes)
            build_asset(args.genome, asset_key, asset_build_package, outfolder, specific_args)
            _LOGGER.info("Finished building asset '{}'".format(asset_key))
        else:
            _LOGGER.warn("Recipe does not exist for asset '{}'".format(asset_key))

    pm.stop_pipeline()


def refgenie_init(genome_config_path, genome_server=DEFAULT_SERVER, config_version=REQ_CFG_VERSION):
    """
    Initialize a genome config file.

    :param str genome_config_path: path to genome configuration file to
        create/initialize
    :param st genome_server: URL for a server
    """

    # Set up default
    rgc = RefGenConf(OrderedDict({
        CFG_VERSION_KEY: config_version,
        CFG_FOLDER_KEY: os.path.dirname(os.path.abspath(genome_config_path)),
        CFG_SERVER_KEY: genome_server,
        CFG_GENOMES_KEY: None
    }))

    _LOGGER.debug("RGC: {}".format(rgc))

    if genome_config_path and not os.path.exists(genome_config_path):
        rgc.write(genome_config_path)
        _LOGGER.info("Wrote new refgenie genome configuration file: {}".format(genome_config_path))
    else:
        _LOGGER.warning("Can't initialize, file exists: {} ".format(genome_config_path))


def _exec_list(rgc, remote, genome):
    if remote:
        pfx = "Remote"
        assemblies, assets = rgc.list_remote(genome=genome)
        recipes = None  # Not implemented
    else:
        pfx = "Local"
        assemblies, assets = rgc.list_local(genome=genome)
        # also get recipes
        recipes = ", ".join(list(asset_build_packages.keys()))

    return pfx, assemblies, assets, recipes


def perm_check_x(file_to_check, message_tag):
    """
    Check X_OK permission on a path, providing according messaging and bool val.

    :param str file_to_check: path to query for permission
    :param str message_tag: context for error message if check fails
    :return bool: os.access(path, X_OK) for the given path
    :raise ValueError: if there's no filepath to check for permission
    """
    if not file_to_check:
        msg = "You must provide a path to {}".format(message_tag)
        _LOGGER.error(msg)
        raise ValueError(msg)
    if not os.access(file_to_check, os.X_OK):
        _LOGGER.error("Insufficient permissions to write to {}: "
                      "{}".format(message_tag, file_to_check))
        return False
    return True


def main():
    """ Primary workflow """

    parser = logmuse.add_logging_options(build_argparser())
    args, remaining_args = parser.parse_known_args()
    global _LOGGER
    _LOGGER = logmuse.logger_via_cli(args)
    logmuse.logger_via_cli(args, name=refgenconf.__name__)
    _LOGGER.info("refgenie {}".format(__version__))
    _LOGGER.debug("Args: {}".format(args))

    if not args.command:
        parser.print_help()
        _LOGGER.error("No command given")
        sys.exit(1)

    gencfg = yacman.select_config(
        args.genome_config, CFG_ENV_VARS,
        check_exist=not args.command == INIT_CMD, on_missing=lambda fp: fp)
    if gencfg is None:
        raise MissingGenomeConfigError(args.genome_config)
    _LOGGER.debug("Determined genome config: {}".format(gencfg))

    if args.command == INIT_CMD:
        _LOGGER.info("Initializing refgenie genome configuration")
        _writeable(os.path.dirname(gencfg), strict_exists=True)
        refgenie_init(gencfg, args.genome_server)
        sys.exit(0)

    rgc = RefGenConf(gencfg)

    if args.command == BUILD_CMD:
        refgenie_build(rgc, args)

    elif args.command == GET_ASSET_CMD:
        _LOGGER.debug("getting asset: '{}/{}'".format(args.genome, args.asset))
        print(" ".join([rgc.get_asset(args.genome, asset) for asset in args.asset]))
        return

    elif args.command == INSERT_CMD:
        if len(args.asset) > 1:
            raise NotImplementedError("Can only add 1 asset at a time")
        else:
            # recast from list to str
            args.asset = args.asset[0]
        refgenie_add(rgc, args)

    elif args.command == PULL_CMD:
        outdir = rgc[CFG_FOLDER_KEY]
        if not os.path.exists(outdir):
            raise MissingFolderError(outdir)
        target = _key_to_name(CFG_FOLDER_KEY)
        if not perm_check_x(outdir, target):
            return
        if not _single_folder_writeable(outdir):
            _LOGGER.error("Insufficient permissions to write to {}: "
                          "{}".format(target, outdir))
            return
        rgc.pull_asset(args.genome, args.asset, gencfg, unpack=not args.no_untar)

    elif args.command in [LIST_LOCAL_CMD, LIST_REMOTE_CMD]:
        pfx, genomes, assets, recipes = _exec_list(rgc, args.command == LIST_REMOTE_CMD, args.genome)
        _LOGGER.info("{} genomes: {}".format(pfx, genomes))
        if args.command != LIST_REMOTE_CMD:  # Not implemented yet
            _LOGGER.info("{} recipes: {}".format(pfx, recipes))
        _LOGGER.info("{} assets:\n{}".format(pfx, assets))

    elif args.command == REMOVE_CMD:
        assets = rgc.list_assets_by_genome(args.genome) if args.asset is None else args.asset
        for asset in assets:
            try:
                rgc.get_asset(args.genome, asset)
            except (MissingAssetError, MissingGenomeError):
                _LOGGER.info("Asset {}/{} does not exist".format(args.genome, asset))
                return
        if len(assets) > 1:
            if not query_yes_no("Remove {} assets for genome: {}?".format(len(assets), args.genome)):
                return
        else:
            query_yes_no("Remove {}/{}?".format(args.genome, assets[0]))
        removed = []
        for asset in assets:
            asset_dir = os.path.abspath(os.path.join(rgc.genome_folder, args.genome, asset))
            asset_archive = asset_dir + ".tar"
            if os.path.isfile(asset_archive):
                os.remove(asset_archive)
                removed.append(asset_archive)
            if os.path.isdir(asset_dir):
                rmtree(asset_dir)
                rgc.remove_assets(args.genome, asset).write()
                removed.append(asset_dir)
        _LOGGER.info("Successfully removed entities:\n- {}".format("\n- ".join(removed)))


def _key_to_name(k):
    return k.replace("_", " ")


def _single_folder_writeable(d):
    return os.access(d, os.W_OK) and os.access(d, os.X_OK)


def _writeable(outdir, strict_exists=False):
    outdir = outdir or "."
    if os.path.exists(outdir):
        return _single_folder_writeable(outdir)
    elif strict_exists:
        raise MissingFolderError(outdir)
    return _writeable(os.path.dirname(outdir), strict_exists)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        _LOGGER.info("Program canceled by user!")
        sys.exit(1)



# This dict provides 'asset packages', which specify recipes (commands) to
# build assets. Each package can produce one or more assets, which are encoded
# as relative paths. The package name is often the same as the asset name but
# it need not be.

# These building recipes should make use of arguments that are auto-populated,
# or user-provided. The auto-populated arguments are:
# - {genome}
# - {asset_outfolder} In addition to these, the recipe should refer in the
#   same way, {var}, to any variables required to be provided, which will be
#   provided via the CLI. These should be listed as 'required_inputs' and
#   will be checked for existence before the commands are executed.

DESC = "description"
ASSET_DESC = "asset_description"
ASSETS = "assets"
PTH = "path"
REQ_IN = "required_inputs"
REQ_ASSETS = "required_assets"
CONT = "container"
CMD_LST = "command_list"


asset_build_packages = {
    "fasta": {
        DESC: "Given a gzipped fasta file, produces fasta, fai, and chrom_sizes assets",
        ASSETS: {
            "fasta": {
                PTH: "fasta/{genome}.fa",
                ASSET_DESC: "Sequences in the FASTA format"
            },
            "fai": {
                PTH: "fasta/{genome}.fa.fai",
                ASSET_DESC: "Indexed fasta file, produced with samtools faidx"
            },
            "chrom_sizes": {
                PTH: "fasta/{genome}.chrom.sizes",
                ASSET_DESC: "Chromosome sizes file"
            }
        },
        REQ_IN: ["fasta"],
        REQ_ASSETS: [],
        CONT: "databio/refgenie",
        CMD_LST: [
            "cp {fasta} {asset_outfolder}/{genome}.fa.gz",
            "gzip -d {asset_outfolder}/{genome}.fa.gz",
            "samtools faidx {asset_outfolder}/{genome}.fa",
            "cut -f 1,2 {asset_outfolder}/{genome}.fa.fai > {asset_outfolder}/{genome}.chrom.sizes",
        ]
    },
    "bowtie2_index": {
        ASSETS: {
            "bowtie2_index": {
                PTH: "bowtie2_index",
                ASSET_DESC: "Genome index for bowtie, produced with bowtie-build"
            }
        },
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        CMD_LST: [
            "bowtie2-build {asset_outfolder}/../fasta/{genome}.fa {asset_outfolder}/{genome}",
            ] 
    },
    "bwa_index": {
        ASSETS: {
            "bwa_index": {
                PTH: "bwa_index",
                ASSET_DESC: "Genome index for Burrows-Wheeler Alignment Tool, produced with bwa index"
            }
        },
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        CMD_LST: [
            "ln -sf ../fasta/{genome}.fa {asset_outfolder}",
            "bwa index {asset_outfolder}/{genome}.fa",
            ] 
    },    
    "hisat2_index": {
        ASSETS: {
            "hisat2_index": {
                PTH: "hisat2_index",
                ASSET_DESC: "Genome index for HISAT2, produced with hisat2-build"
            }
        },     
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        CMD_LST: [
            "hisat2-build {asset_outfolder}/../fasta/{genome}.fa {asset_outfolder}/{genome}"
            ] 
    },
    "bismark_bt2_index": {
        DESC: "The fasta asset must be built first for this to work.",
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        ASSETS: {
            "bismark_bt2_index": {
                PTH: "bismark_bt2_index",
                ASSET_DESC: "Genome index for Bisulfite-Seq applications, produced by bismark_genome_preparation using bowtie2"
            }
        },       
        CMD_LST: [
            "ln -sf ../fasta/{genome}.fa {asset_outfolder}",
            "bismark_genome_preparation --bowtie2 {asset_outfolder}"
            ] 
    },
    "bismark_bt1_index": {
        DESC: "The fasta asset must be built first for this to work.",
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        ASSETS: {
            "bismark_bt1_index": {
                PTH: "bismark_bt1_index",
                ASSET_DESC: "Genome index for Bisulfite-Seq applications, produced by bismark_genome_preparation using bowtie"
            }
        },       
        CMD_LST: [
            "ln -sf ../fasta/{genome}.fa {asset_outfolder}",
            "bismark_genome_preparation {asset_outfolder}"
            ] 
    },  
    "kallisto_index": {
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        ASSETS: {
            "kallisto_index": {
                PTH: "kallisto_index",
                ASSET_DESC: "Genome index for kallisto, produced with kallisto index"
            }
        },
        CMD_LST: [
            "kallisto index -i {asset_outfolder}/{genome}_kallisto_index.idx {asset_outfolder}/../fasta/{genome}.fa"
            ] 
    },
    "salmon_index": {
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "combinelab/salmon",
        ASSETS: {
            "salmon_index": {
                PTH: "salmon_index",
                ASSET_DESC: "Transcriptome index for salmon, produced with salmon index"
            }
        },
        CMD_LST: [
            "salmon index -k 31 -i {asset_outfolder} -t {asset_outfolder}/../fasta/{genome}.fa"
            ] 
    },
    "epilog_index": {
        REQ_IN: ["context"],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        ASSETS: {
            "epilog_index": {
                PTH: "epilog_index",
                ASSET_DESC: "Genome index for CpG sites, produced by the epilog DNA methylation caller"
            }
        },
        CMD_LST: [
            "epilog index -i {asset_outfolder}/../fasta/{genome}.fa -o {asset_outfolder}/{genome}_{context}.tsv -s {context} -t"
            ] 
    },
    "star_index": {
        REQ_IN: [],
        REQ_ASSETS: ["fasta"],
        CONT: "databio/refgenie",
        ASSETS: {
            "star_index": {
                PTH: "star_index",
                ASSET_DESC: "Genome index for STAR RNA-seq aligner, produced with STAR --runMode genomeGenerate"
            }
        },
        CMD_LST: [
            "mkdir -p {asset_outfolder}",
            "STAR --runThreadN 16 --runMode genomeGenerate --genomeDir {asset_outfolder} --genomeFastaFiles {asset_outfolder}/../fasta/{genome}.fa "
            ]
    },
    "gencode_gtf": {
        DESC: "Given a GTF file (must be downloaded), create a GTF annotation asset.  GTF provides access to all annotated transcripts which make up an Ensembl gene set.",
        REQ_IN: ["gtf"],
        REQ_ASSETS: [],
        CONT: "databio/refgenie",
        ASSETS: {
            "gencode_gtf": {
                PTH: "gencode_gtf/{genome}.gtf.gz",
                ASSET_DESC: "Gencode GTF provides access to all gene features with high accuracy based on biological evidence."
            }
        },
        CMD_LST: [
            "cp {gtf} {asset_outfolder}/{genome}.gtf.gz"
            ] 
    },
    "ensembl_gtf": {
        DESC: "Given a Ensembl GTF file (must be downloaded), create Ensembl GTF, TSS, and gene body annotation assets.",
        REQ_IN: ["gtf"],
        REQ_ASSETS: [],
        CONT: "databio/refgenie",
        ASSETS: {
            "ensembl_gtf": {
                PTH: "ensembl_gtf/{genome}.gtf.gz",
                ASSET_DESC: "Ensembl GTF provides access to all annotated transcripts which make up an Ensembl gene set"
            },
            "ensembl_tss": {
                PTH: "ensembl_gtf/{genome}_ensembl_TSS.bed",
                ASSET_DESC: "Ensembl transcription start sites (TSSs) derived from an Ensembl GTF annotation."
            },
            "ensembl_gene_body": {
                PTH: "ensembl_gtf/{genome}_ensembl_gene_body.bed",
                ASSET_DESC: "Ensembl gene bodies derived from an Ensembl GTF annotation."
            }
        },
        CMD_LST: [
            "cp {gtf} {asset_outfolder}/{genome}.gtf.gz",
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | grep 'exon_number \"1\";' | sed 's/^/chr/' | awk -v OFS='\t' '{{print $1, $4, $5, $20, $14, $7}}' | sed 's/\";//g' | sed 's/\"//g' | awk '{{if($6==\"+\"){{print $1\"\t\"$2+20\"\t\"$2+120\"\t\"$4\"\t\"$5\"\t\"$6}}else{{print $1\"\t\"$3-120\"\t\"$3-20\"\t\"$4\"\t\"$5\"\t\"$6}}}}' | LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_ensembl_TSS.bed",
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | awk '$3 == \"gene\"' | sed 's/^/chr/' | awk -v OFS='\t' '{{print $1,$4,$5,$14,$6,$7}}' | sed 's/\";//g' | sed 's/\"//g' | awk '$4!=\"Metazoa_SRP\"' | awk '$4!=\"U3\"' | awk '$4!=\"7SK\"'  | awk '($3-$2)>200' | awk '{{if($6==\"+\"){{print $1\"\t\"$2+500\"\t\"$3\"\t\"$4\"\t\"$5\"\t\"$6}}else{{print $1\"\t\"$2\"\t\"$3-500\"\t\"$4\"\t\"$5\"\t\"$6}}}}' | awk '$3>$2' | LC_COLLATE=C sort -k4 -u > {asset_outfolder}/{genome}_ensembl_gene_body.bed"
            ] 
    },
    "ensembl_rb": {
        DESC: "Given a GFF regulatory build file (must be downloaded), create a regulatory annotation asset.",
        REQ_IN: ["gff"],
        REQ_ASSETS: [],
        CONT: "databio/refgenie",
        ASSETS: {
            "ensembl_rb": {
                PTH: "ensembl_rb/{genome}.gff.gz",
                ASSET_DESC: "A genome-wide set of regions that are likely to be involved in gene regulation."
            }
        },
        CMD_LST: [
            "cp {gff} {asset_outfolder}/{genome}.gff.gz"
            ] 
    },
    "refgene_anno": {
        DESC: "Given a refGene file (must be downloaded), create gene, TSS, exon, intron, and premature mRNA annotation assets.",
        REQ_IN: ["refgene"],
        REQ_ASSETS: [],
        CONT: "databio/refgenie",
        ASSETS: {
            "refgene_anno": {
                PTH: "refgene_anno/{genome}_refGene.txt.gz",
                ASSET_DESC: "Specifies known human protein-coding and non-protein-coding genes taken from the NCBI RNA reference sequences collection (RefSeq)."
            },
            "refgene_tss": {
                PTH: "refgene_anno/{genome}_TSS.bed",
                ASSET_DESC: "Transcription start sites (TSSs) derived from the refGene annotation asset."
            },
            "refgene_exon": {
                PTH: "refgene_anno/{genome}_exons.bed",
                ASSET_DESC: "Exons derived from the refGene annotation asset."
            },
            "refgene_intron": {
                PTH: "refgene_anno/{genome}_introns.bed",
                ASSET_DESC: "Introns derived from the refGene annotation asset."
            },
            "refgene_pre_mRNA": {
                PTH: "refgene_anno/{genome}_pre-mRNA.bed",
                ASSET_DESC: "Premature mRNA from complete genes derived from the refGene annotation asset"
            }
        },
        CMD_LST: [
            "cp {refgene} {asset_outfolder}/{genome}_refGene.txt.gz",
            "gzip -dc {asset_outfolder}/../refgene_anno/{genome}_refGene.txt.gz | awk '{{if($4==\"+\"){{print $3\"\t\"$5\"\t\"$5\"\t\"$13\"\t.\t\"$4}}else{{print $3\"\t\"$6\"\t\"$6\"\t\"$13\"\t.\t\"$4}}}}' | LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_TSS.bed",
            "gzip -dc {asset_outfolder}/../refgene_anno/{genome}_refGene.txt.gz | awk -v OFS='\t' '$9>1' | awk -v OFS='\t' '{{ n = split($10, a, \",\"); split($11, b, \",\"); for(i=1; i<n; ++i) print $3, a[i], b[i], $13, i, $4 }}' | awk -v OFS='\t' '$6==\"+\" && $5!=1 {{print $0}} $6==\"-\" {{print $0}}' | awk '$4!=prev4 && prev6==\"-\" {{prev4=$4; prev6=$6; delete line[NR-1]; idx-=1}} {{line[++idx]=$0; prev4=$4; prev6=$6}} END {{for (x=1; x<=idx; x++) print line[x]}}' | LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_exons.bed",
            "gzip -dc {asset_outfolder}/../refgene_anno/{genome}_refGene.txt.gz | awk -v OFS='\t' '$9>1' | awk -F'\t' '{{ exonCount=int($9);split($10,exonStarts,\"[,]\"); split($11,exonEnds,\"[,]\"); for(i=1;i<exonCount;i++) {{printf(\"%s\\t%s\\t%s\\t%s\\t%d\\t%s\\n\",$3,exonEnds[i],exonStarts[i+1],$13,($3==\"+\"?i:exonCount-i),$4);}}}}' | LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_introns.bed",
            "gzip -dc {asset_outfolder}/../refgene_anno/{genome}_refGene.txt.gz | grep 'cmpl' | awk  '{{print $3\"\t\"$5\"\t\"$6\"\t\"$13\"\t.\t\"$4}}' | LC_COLLATE=C sort -k1,1 -k2,2n -u >  {asset_outfolder}/{genome}_pre-mRNA.bed"
            ]
    },
    "feat_annotation": {
        DESC: "Using a Ensembl GTF annotation asset and an Ensembl regulatory build annotation asset, create a combined genomic feature annotation asset.",
        ASSETS: {
            "feat_annotation": {
                PTH: "feat_annotation/{genome}_annotations.bed.gz",
                ASSET_DESC: "Genomic features including enhancers, promoters, promoter flanking, 5' UTR, 3' UTR, exons, and introns derived from Ensembl GTF and Ensembl regulatory build (rb) assets."
            }
        },
        REQ_IN: [],
        REQ_ASSETS: ["ensembl_gtf", "ensembl_rb"],
        CONT: "databio/refgenie",
        CMD_LST: [
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | awk '$3==\"exon\"' | grep -v 'pseudogene' | awk -v OFS='\t' '{{print \"chr\"$1, $4-1, $5, \"Exon\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3n -u > {asset_outfolder}/{genome}_exons.bed",
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | awk '$3==\"exon\"' | grep -v 'pseudogene' | awk -v OFS='\t' '{{ split($20, a, \"\\\"\"); print \"chr\"$1, $4-1, $5, a[2], $6, $7}}' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3n -u | awk 'seen[$4]++ && seen[$4] > 1' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3nr | env LC_COLLATE=C sort -k1,1 -k2,2n -u | env LC_COLLATE=C sort -k1,1 -k3,3n -u | awk -v OFS='\t' '{{if($4==prev4){{new2=prev3+1;}} {{prev4=$4; prev3=$3; print $1, new2, $2-1, \"Intron\", $5, $6}}}}' | awk -F'\t' '$2' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_introns.bed",
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | awk '$3==\"three_prime_utr\"' | grep -v 'pseudogene' | awk -v OFS='\t' '{{print \"chr\"$1, $4-1, $5, \"3\'\\\'\' UTR\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_3utr.bed",
            "gzip -dc {asset_outfolder}/../ensembl_gtf/{genome}.gtf.gz | awk '$3==\"five_prime_utr\"' | grep -v 'pseudogene' | awk -v OFS='\t' '{{print \"chr\"$1, $4-1, $5, \"5\'\\\'\' UTR\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -u > {asset_outfolder}/{genome}_5utr.bed",
            "gzip -dc {asset_outfolder}/../ensembl_rb/{genome}.gff.gz | awk '$3==\"promoter\"' | awk -v OFS='\t' '{{print \"chr\"$1, $4, $5, \"Promoter\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3n -u > {asset_outfolder}/{genome}_promoter.bed",
            "gzip -dc {asset_outfolder}/../ensembl_rb/{genome}.gff.gz | awk '$3==\"promoter_flanking_region\"' | awk -v OFS='\t' '{{print \"chr\"$1, $4, $5, \"Promoter Flanking Region\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3n -u > {asset_outfolder}/{genome}_promoter_flanking.bed",
            "gzip -dc {asset_outfolder}/../ensembl_rb/{genome}.gff.gz | awk '$3==\"enhancer\"' | awk -v OFS='\t' '{{print \"chr\"$1, $4, $5, \"Enhancer\", $6, $7}}' | awk '$2<$3' | env LC_COLLATE=C sort -k1,1 -k2,2n -k3,3n -u > {asset_outfolder}/{genome}_enhancer.bed",
            "cat {asset_outfolder}/{genome}_enhancer.bed {asset_outfolder}/{genome}_promoter.bed {asset_outfolder}/{genome}_promoter_flanking.bed {asset_outfolder}/{genome}_5utr.bed {asset_outfolder}/{genome}_3utr.bed {asset_outfolder}/{genome}_exons.bed {asset_outfolder}/{genome}_introns.bed | awk -F'\t' '!seen[$1, $2, $3]++' | env LC_COLLATE=C sort -k4d -k1.4,1V -k2,2n -s > {asset_outfolder}/{genome}_annotations.bed",
            "rm -f {asset_outfolder}/{genome}_enhancer.bed {asset_outfolder}/{genome}_promoter.bed {asset_outfolder}/{genome}_promoter_flanking.bed {asset_outfolder}/{genome}_5utr.bed {asset_outfolder}/{genome}_3utr.bed {asset_outfolder}/{genome}_exons.bed {asset_outfolder}/{genome}_introns.bed",
            "gzip {asset_outfolder}/{genome}_annotations.bed"
            ]
    }
}


